from typing import Tuple, Type

from data_gateway_sdk.protocols.device_packet.device_packet import DevicePacketTyped
from data_gateway_sdk.protocols.water5_v_nero.water5_device_cold_reset import Water5DeviceColdResetData
from data_gateway_sdk.protocols.water5_v_nero.water5_device_daily import Water5DeviceDailyData
from data_gateway_sdk.protocols.water5_v_nero.water5_device_info_ch1 import Water5DeviceInfoCh1Data
from data_gateway_sdk.protocols.water5_v_nero.water5_device_info_ch2 import Water5DeviceInfoCh2Data
from data_gateway_sdk.protocols.water5_v_nero.water5_device_magnet import Water5DeviceMagnetData
from data_gateway_sdk.protocols.water5_v_nero.water5_device_warm_reset_v_fluo_a import Water5DeviceWarmResetVFluoAData
from data_gateway_sdk.protocols.water5_v_nero.water5_device_warm_reset_v_fluo_freescale import Water5DeviceWarmResetVFluoFreescaleData
from data_gateway_sdk.protocols.water5_v_nero.water5_device_warm_reset_v_fluo_stm import Water5DeviceWarmResetVFluoStmData
from data_gateway_sdk.protocols.water5_v_nero.water5_device_warm_reset_v_jupiter_freescale import Water5DeviceWarmResetVJupiterFreescaleData
from data_gateway_sdk.protocols.water5_v_nero.water5_device_warm_reset_v_jupiter_stm import Water5DeviceWarmResetVJupiterStmData
from data_gateway_sdk.protocols.water5_v_nero.water5_device_warm_reset_v_metano_a import \
    Water5DeviceWarmResetVMetanoAData
from data_gateway_sdk.protocols.water5_v_nero.water5_device_weekly_ch1 import Water5DeviceWeeklyCh1Data
from data_gateway_sdk.protocols.water5_v_nero.water5_device_weekly_ch1_impulse import Water5DeviceWeeklyCh1ImpulseData
from data_gateway_sdk.protocols.water5_v_nero.water5_device_weekly_ch1_impulse_v2 import Water5DeviceWeeklyCh1ImpulseV2Data
from data_gateway_sdk.protocols.water5_v_nero.water5_device_weekly_ch2 import Water5DeviceWeeklyCh2Data
from data_gateway_sdk.protocols.water5_v_nero.water5_device_weekly_ch2_impulse import Water5DeviceWeeklyCh2ImpulseData
from data_gateway_sdk.protocols.water5_v_nero.water5_device_weekly_ch2_impulse_v2 import Water5DeviceWeeklyCh2ImpulseV2Data
from data_gateway_sdk.utils.packet import Packet


class Water5NeroV0DevicePacket(DevicePacketTyped):
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        Water5DeviceDailyData,
        Water5DeviceWeeklyCh1Data,
        Water5DeviceWeeklyCh2Data,
        Water5DeviceInfoCh1Data,
        Water5DeviceInfoCh2Data,
        Water5DeviceMagnetData,
        Water5DeviceColdResetData,
    )


class Water5GasV0DevicePacket(DevicePacketTyped):
    # MM215. DAILY, WEEKLY, INFO, WARM_RESET
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        Water5DeviceWarmResetVMetanoAData,
        Water5DeviceDailyData,
        Water5DeviceWeeklyCh1Data,
        Water5DeviceInfoCh1Data,
    )


class Water5FluoAV0DevicePacket(DevicePacketTyped):
    # MM217. WEEKLY, INFO, COLD_RESET, WARM_RESET
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        Water5DeviceWeeklyCh1Data,
        Water5DeviceInfoCh1Data,
        Water5DeviceColdResetData,
        Water5DeviceWarmResetVFluoAData,
    )


class Water5FluoSV0DevicePacket(DevicePacketTyped):
    # MM217. WEEKLY, INFO, COLD_RESET, WARM_RESET
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        Water5DeviceDailyData,
        Water5DeviceWeeklyCh1Data,
        Water5DeviceColdResetData,
    )


class Water5FluoFreeScaleV0DevicePacket(DevicePacketTyped):
    # MM214. CH1. WEEKLY, INFO, COLD_RESET, WARM_RESET, COMMAND
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        Water5DeviceDailyData,
        Water5DeviceWeeklyCh1Data,
        Water5DeviceInfoCh1Data,
        Water5DeviceMagnetData,
        Water5DeviceWarmResetVFluoFreescaleData,
    )  # noqa: E501


class Water5FluoLoraV0DevicePacket(DevicePacketTyped):  #
    # FLUO LORA. MM224. CH_1 - прямой, СH_2 - обратный. WEEKLY
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        Water5DeviceWeeklyCh1Data,
        Water5DeviceWeeklyCh2Data,
    )


class Water5JupiterLoraV0DevicePacket(DevicePacketTyped):  #
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        Water5DeviceWeeklyCh1ImpulseData,
        Water5DeviceWeeklyCh2ImpulseData,
    )
    # JUPITER LORA. MM216. CH_1 - прямой, СH_2 - обратный. WEEKLY.  АБСОЛЮТНО ТАКОЙ ЖЕ КАК И FLUO LORA (MM224)


class Water5FluoSTMV0DevicePacket(DevicePacketTyped):
    # FLUO STM. MM223 - huashun, MM222 - vnlink. WEEKLY, INFO, COMMAND, WARM_RESET
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        Water5DeviceWeeklyCh1Data,
        Water5DeviceInfoCh1Data,
        Water5DeviceMagnetData,
        Water5DeviceWarmResetVFluoStmData,
    )


class Water5JupiterSTMV0DevicePacket(DevicePacketTyped):
    # Jupiter STM. MM212. CH_1, СH_2. WEEKLY, WARM_RESET
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        Water5DeviceWeeklyCh1ImpulseData,
        Water5DeviceWeeklyCh2ImpulseData,
        Water5DeviceWarmResetVJupiterStmData,
    )


class Water5JupiterFreeScaleV0DevicePacket(DevicePacketTyped):
    # Jupiter Free Scale. MM212, CH_1, СH_2, WEEKLY, WARM_RESET, 24 bits
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = (
        Water5DeviceWeeklyCh1ImpulseV2Data,
        Water5DeviceWeeklyCh2ImpulseV2Data,
        Water5DeviceWarmResetVJupiterFreescaleData,
    )



from datetime import timedelta, datetime, tzinfo, time
from typing import Any, List, Dict
from decimal import Decimal

from data_aggregator_sdk.constants.enums import IntegrationV0MessageEvent, JournalDataType, SensorType
from data_aggregator_sdk.integration_message import IntegrationV0MessageData, ResourceType, CounterType, IntegrationV0MessageConsumption, IntegrationV0MessageSensor

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.days_ago_calculation import days_ago_calculation
from data_gateway_sdk.utils.packet import Packet


# PACKET (384 bits)   smpm_ul_device_heat_group_meter_daily
#
# RESULT int:        36796229577608994159379990843197031097932211333
# RESULT bin:  MSB   00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000110 01110010 00000000 00000000 00000000 00000110 01110010 00000000 00000000 00000000 00000110 01110010 00000000 00000000 00000000 00000110 01110010 00000000 01100100 10000101   LSB              # noqa: E501
# RESULT hex:  LE    85 64 00 72 06 00 00 00 72 06 00 00 00 72 06 00 00 00 72 06 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00
#
# name                                                type       size  value(int)                                                                                                                                                                                                                                                                                                                                                                                        data(bits)     # noqa: E501
# --------------------------------------------------  ---------  ----  ----------  ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------     # noqa: E501
# packet_type_id.0.VAL                                u7            7           5                                                                                                                                                                                                                                                                                                                                                                                           0000101     # noqa: E501
# packet_type_id.0.DFF                                bool          1           1                                                                                                                                                                                                                                                                                                                                                                                          1            # noqa: E501
# packet_type_id.1.VAL                                u2            2           0                                                                                                                                                                                                                                                                                                                                                                                        00             # noqa: E501
# packet_type_id.1.DFF                                bool          1           1                                                                                                                                                                                                                                                                                                                                                                                       1               # noqa: E501
# packet_type_id.2.VAL                                u2            2           0                                                                                                                                                                                                                                                                                                                                                                                     00                # noqa: E501
# packet_type_id.2.DFF                                bool          1           1                                                                                                                                                                                                                                                                                                                                                                                    1                  # noqa: E501
# packet_type_id.3.VAL                                u2            2           1                                                                                                                                                                                                                                                                                                                                                                                  01                   # noqa: E501
# packet_type_id.3.DFF                                bool          1           0                                                                                                                                                                                                                                                                                                                                                                                 0                     # noqa: E501
# days_ago                                            timedelta     6           0                                                                                                                                                                                                                                                                                                                                                                           000000                      # noqa: E501
# value_ch1                                           uf40p3       40        3300                                                                                                                                                                                                                                                                                                                                   0000000000000000000000000000110011100100                            # noqa: E501
# value_ch2                                           uf40p3       40        3300                                                                                                                                                                                                                                                                                           0000000000000000000000000000110011100100                                                                    # noqa: E501
# value_ch3                                           uf40p3       40        3300                                                                                                                                                                                                                                                   0000000000000000000000000000110011100100                                                                                                            # noqa: E501
# value_ch4                                           uf40p3       40        3300                                                                                                                                                                                                           0000000000000000000000000000110011100100                                                                                                                                                    # noqa: E501
# uptime_ch1                                          timedelta    22           0                                                                                                                                                                                     0000000000000000000000                                                                                                                                                                                            # noqa: E501
# uptime_ch2                                          timedelta    22           0                                                                                                                                                               0000000000000000000000                                                                                                                                                                                                                  # noqa: E501
# uptime_ch3                                          timedelta    22           0                                                                                                                                         0000000000000000000000                                                                                                                                                                                                                                        # noqa: E501
# uptime_ch4                                          timedelta    22           0                                                                                                                   0000000000000000000000                                                                                                                                                                                                                                                              # noqa: E501
# event_device_power_supply_lost                      bool          1           0                                                                                                                  0                                                                                                                                                                                                                                                                                    # noqa: E501
# event_device_power_supply_enabled                   bool          1           0                                                                                                                 0                                                                                                                                                                                                                                                                                     # noqa: E501
# event_device_common_settings_changed                bool          1           0                                                                                                                0                                                                                                                                                                                                                                                                                      # noqa: E501
# event_device_measure_channels_settings_changed      bool          1           0                                                                                                               0                                                                                                                                                                                                                                                                                       # noqa: E501
# event_device_digital_i_o_settings_changed           bool          1           0                                                                                                              0                                                                                                                                                                                                                                                                                        # noqa: E501
# event_device_datetime_changed                       bool          1           0                                                                                                             0                                                                                                                                                                                                                                                                                         # noqa: E501
# event_device_ethernet_settings_changed              bool          1           0                                                                                                            0                                                                                                                                                                                                                                                                                          # noqa: E501
# event_device_system_settings_changed1               bool          1           0                                                                                                           0                                                                                                                                                                                                                                                                                           # noqa: E501
# event_device_system_settings_changed2               bool          1           0                                                                                                          0                                                                                                                                                                                                                                                                                            # noqa: E501
# event_device_system_settings_changed3               bool          1           0                                                                                                         0                                                                                                                                                                                                                                                                                             # noqa: E501
# event_device_system_settings_changed4               bool          1           0                                                                                                        0                                                                                                                                                                                                                                                                                              # noqa: E501
# event_device_alarm_digital_input1                   bool          1           0                                                                                                       0                                                                                                                                                                                                                                                                                               # noqa: E501
# event_device_alarm_overrate_digital_input1          bool          1           0                                                                                                      0                                                                                                                                                                                                                                                                                                # noqa: E501
# event_device_alarm_overrate_digital_input2          bool          1           0                                                                                                     0                                                                                                                                                                                                                                                                                                 # noqa: E501
# event_device_alarm_underrate_digital_input1         bool          1           0                                                                                                    0                                                                                                                                                                                                                                                                                                  # noqa: E501
# event_device_alarm_underrate_digital_input2         bool          1           0                                                                                                   0                                                                                                                                                                                                                                                                                                   # noqa: E501
# event_device_alarm_t_high_digital_input1            bool          1           0                                                                                                  0
# event_device_alarm_t_high_digital_input2            bool          1           0                                                                                                 0
# event_device_alarm_t_low_digital_input1             bool          1           0                                                                                                0
# event_device_alarm_t_low_digital_input2             bool          1           0                                                                                               0
# event_device_alarm_delta_t_high_digital_input1      bool          1           0                                                                                              0
# event_device_alarm_delta_t_high_digital_input2      bool          1           0                                                                                             0
# event_device_alarm_delta_t_low_digital_input1       bool          1           0                                                                                            0
# event_device_alarm_delta_t_low_digital_input2       bool          1           0                                                                                           0
# event_device_alarm_power_high_digital_input1        bool          1           0                                                                                          0
# event_device_alarm_power_high_digital_input2        bool          1           0                                                                                         0
# event_device_alarm_power_low_digital_input1         bool          1           0                                                                                        0
# event_device_alarm_power_low_digital_input2         bool          1           0                                                                                       0
# event_ch1_circuit_break_t_sensor_1                  bool          1           0                                                                                      0
# event_ch1_circuit_break_t_sensor_2                  bool          1           0                                                                                     0
# event_ch1_circuit_break_t_sensor_3                  bool          1           0                                                                                    0
# event_ch1_error_delta_t                             bool          1           0                                                                                   0
# event_ch1_consumption_lower_g_min_channel_rate_1    bool          1           0                                                                                  0
# event_ch1_consumption_lower_g_min_channel_rate_2    bool          1           0                                                                                 0
# event_ch1_consumption_lower_g_min_channel_rate_3    bool          1           0                                                                                0
# event_ch1_consumption_lower_g_max_channel_rate_1    bool          1           0                                                                               0
# event_ch1_consumption_lower_g_max_channel_rate_2    bool          1           0                                                                              0
# event_ch1_consumption_lower_g_max_channel_rate_3    bool          1           0                                                                             0
# event_ch1_no_coolant_channel_rate_1                 bool          1           0                                                                            0
# event_ch1_no_coolant_channel_rate_2                 bool          1           0                                                                           0
# event_ch1_no_coolant_channel_rate_3                 bool          1           0                                                                          0
# event_ch1_circuit_break_stimulation_channel_rate_1  bool          1           0                                                                         0
# event_ch1_circuit_break_stimulation_channel_rate_2  bool          1           0                                                                        0
# event_ch1_circuit_break_pressure_sensor_1           bool          1           0                                                                       0
# event_ch1_circuit_break_pressure_sensor_2           bool          1           0                                                                      0
# event_ch1_circuit_break_pressure_sensor_3           bool          1           0                                                                     0
# event_ch1_reverse                                   bool          1           0                                                                    0
# event_ch2_circuit_break_t_sensor_1                  bool          1           0                                                                   0
# event_ch2_circuit_break_t_sensor_2                  bool          1           0                                                                  0
# event_ch2_circuit_break_t_sensor_3                  bool          1           0                                                                 0
# event_ch2_error_delta_t                             bool          1           0                                                                0
# event_ch2_consumption_lower_g_min_channel_rate_1    bool          1           0                                                               0
# event_ch2_consumption_lower_g_min_channel_rate_2    bool          1           0                                                              0
# event_ch2_consumption_lower_g_min_channel_rate_3    bool          1           0                                                             0
# event_ch2_consumption_lower_g_max_channel_rate_1    bool          1           0                                                            0
# event_ch2_consumption_lower_g_max_channel_rate_2    bool          1           0                                                           0
# event_ch2_consumption_lower_g_max_channel_rate_3    bool          1           0                                                          0
# event_ch2_no_coolant_channel_rate_1                 bool          1           0                                                         0
# event_ch2_no_coolant_channel_rate_2                 bool          1           0                                                        0
# event_ch2_no_coolant_channel_rate_3                 bool          1           0                                                       0
# event_ch2_circuit_break_stimulation_channel_rate_1  bool          1           0                                                      0
# event_ch2_circuit_break_stimulation_channel_rate_2  bool          1           0                                                     0
# event_ch2_circuit_break_pressure_sensor_1           bool          1           0                                                    0
# event_ch2_circuit_break_pressure_sensor_2           bool          1           0                                                   0
# event_ch2_circuit_break_pressure_sensor_3           bool          1           0                                                  0
# event_ch2_reverse                                   bool          1           0                                                 0
# event_ch3_circuit_break_t_sensor_1                  bool          1           0                                                0
# event_ch3_circuit_break_t_sensor_2                  bool          1           0                                               0
# event_ch3_circuit_break_t_sensor_3                  bool          1           0                                              0
# event_ch3_error_delta_t                             bool          1           0                                             0
# event_ch3_consumption_lower_g_min_channel_rate_1    bool          1           0                                            0
# event_ch3_consumption_lower_g_min_channel_rate_2    bool          1           0                                           0
# event_ch3_consumption_lower_g_min_channel_rate_3    bool          1           0                                          0
# event_ch3_consumption_lower_g_max_channel_rate_1    bool          1           0                                         0
# event_ch3_consumption_lower_g_max_channel_rate_2    bool          1           0                                        0
# event_ch3_consumption_lower_g_max_channel_rate_3    bool          1           0                                       0
# event_ch3_no_coolant_channel_rate_1                 bool          1           0                                      0
# event_ch3_no_coolant_channel_rate_2                 bool          1           0                                     0
# event_ch3_no_coolant_channel_rate_3                 bool          1           0                                    0
# event_ch3_circuit_break_stimulation_channel_rate_1  bool          1           0                                   0
# event_ch3_circuit_break_stimulation_channel_rate_2  bool          1           0                                  0
# event_ch3_circuit_break_pressure_sensor_1           bool          1           0                                 0
# event_ch3_circuit_break_pressure_sensor_2           bool          1           0                                0
# event_ch3_circuit_break_pressure_sensor_3           bool          1           0                               0
# event_ch3_reverse                                   bool          1           0                              0
# event_ch4_circuit_break_t_sensor_1                  bool          1           0                             0
# event_ch4_circuit_break_t_sensor_2                  bool          1           0                            0
# event_ch4_circuit_break_t_sensor_3                  bool          1           0                           0
# event_ch4_error_delta_t                             bool          1           0                          0
# event_ch4_consumption_lower_g_min_channel_rate_1    bool          1           0                         0
# event_ch4_consumption_lower_g_min_channel_rate_2    bool          1           0                        0
# event_ch4_consumption_lower_g_min_channel_rate_3    bool          1           0                       0
# event_ch4_consumption_lower_g_max_channel_rate_1    bool          1           0                      0
# event_ch4_consumption_lower_g_max_channel_rate_2    bool          1           0                     0
# event_ch4_consumption_lower_g_max_channel_rate_3    bool          1           0                    0
# event_ch4_no_coolant_channel_rate_1                 bool          1           0                   0
# event_ch4_no_coolant_channel_rate_2                 bool          1           0                  0
# event_ch4_no_coolant_channel_rate_3                 bool          1           0                 0
# event_ch4_circuit_break_stimulation_channel_rate_1  bool          1           0                0
# event_ch4_circuit_break_stimulation_channel_rate_2  bool          1           0               0
# event_ch4_circuit_break_pressure_sensor_1           bool          1           0              0
# event_ch4_circuit_break_pressure_sensor_2           bool          1           0             0
# event_ch4_circuit_break_pressure_sensor_3           bool          1           0            0
# event_ch4_reverse                                   bool          1           0           0
# RESERVED                                            u9            9           0  000000000


class SmpmUlDeviceHeatGroupMeterDailyData(Packet):
    days_ago: timedelta
    value_ch1: float
    value_ch2: float
    value_ch3: float
    value_ch4: float
    uptime_ch1: timedelta
    uptime_ch2: timedelta
    uptime_ch3: timedelta
    uptime_ch4: timedelta
    event_device_power_supply_lost: bool
    event_device_power_supply_enabled: bool
    event_device_common_settings_changed: bool
    event_device_measure_channels_settings_changed: bool
    event_device_digital_i_o_settings_changed: bool
    event_device_datetime_changed: bool
    event_device_ethernet_settings_changed: bool
    event_device_system_settings_changed1: bool
    event_device_system_settings_changed2: bool
    event_device_system_settings_changed3: bool
    event_device_system_settings_changed4: bool
    event_device_alarm_digital_input1: bool
    event_device_alarm_overrate_digital_input1: bool
    event_device_alarm_overrate_digital_input2: bool
    event_device_alarm_underrate_digital_input1: bool
    event_device_alarm_underrate_digital_input2: bool
    event_device_alarm_t_high_digital_input1: bool
    event_device_alarm_t_high_digital_input2: bool
    event_device_alarm_t_low_digital_input1: bool
    event_device_alarm_t_low_digital_input2: bool
    event_device_alarm_delta_t_high_digital_input1: bool
    event_device_alarm_delta_t_high_digital_input2: bool
    event_device_alarm_delta_t_low_digital_input1: bool
    event_device_alarm_delta_t_low_digital_input2: bool
    event_device_alarm_power_high_digital_input1: bool
    event_device_alarm_power_high_digital_input2: bool
    event_device_alarm_power_low_digital_input1: bool
    event_device_alarm_power_low_digital_input2: bool
    event_ch1_circuit_break_t_sensor_1: bool
    event_ch1_circuit_break_t_sensor_2: bool
    event_ch1_circuit_break_t_sensor_3: bool
    event_ch1_error_delta_t: bool
    event_ch1_consumption_lower_g_min_channel_rate_1: bool
    event_ch1_consumption_lower_g_min_channel_rate_2: bool
    event_ch1_consumption_lower_g_min_channel_rate_3: bool
    event_ch1_consumption_lower_g_max_channel_rate_1: bool
    event_ch1_consumption_lower_g_max_channel_rate_2: bool
    event_ch1_consumption_lower_g_max_channel_rate_3: bool
    event_ch1_no_coolant_channel_rate_1: bool
    event_ch1_no_coolant_channel_rate_2: bool
    event_ch1_no_coolant_channel_rate_3: bool
    event_ch1_circuit_break_stimulation_channel_rate_1: bool
    event_ch1_circuit_break_stimulation_channel_rate_2: bool
    event_ch1_circuit_break_pressure_sensor_1: bool
    event_ch1_circuit_break_pressure_sensor_2: bool
    event_ch1_circuit_break_pressure_sensor_3: bool
    event_ch1_reverse: bool
    event_ch2_circuit_break_t_sensor_1: bool
    event_ch2_circuit_break_t_sensor_2: bool
    event_ch2_circuit_break_t_sensor_3: bool
    event_ch2_error_delta_t: bool
    event_ch2_consumption_lower_g_min_channel_rate_1: bool
    event_ch2_consumption_lower_g_min_channel_rate_2: bool
    event_ch2_consumption_lower_g_min_channel_rate_3: bool
    event_ch2_consumption_lower_g_max_channel_rate_1: bool
    event_ch2_consumption_lower_g_max_channel_rate_2: bool
    event_ch2_consumption_lower_g_max_channel_rate_3: bool
    event_ch2_no_coolant_channel_rate_1: bool
    event_ch2_no_coolant_channel_rate_2: bool
    event_ch2_no_coolant_channel_rate_3: bool
    event_ch2_circuit_break_stimulation_channel_rate_1: bool
    event_ch2_circuit_break_stimulation_channel_rate_2: bool
    event_ch2_circuit_break_pressure_sensor_1: bool
    event_ch2_circuit_break_pressure_sensor_2: bool
    event_ch2_circuit_break_pressure_sensor_3: bool
    event_ch2_reverse: bool
    event_ch3_circuit_break_t_sensor_1: bool
    event_ch3_circuit_break_t_sensor_2: bool
    event_ch3_circuit_break_t_sensor_3: bool
    event_ch3_error_delta_t: bool
    event_ch3_consumption_lower_g_min_channel_rate_1: bool
    event_ch3_consumption_lower_g_min_channel_rate_2: bool
    event_ch3_consumption_lower_g_min_channel_rate_3: bool
    event_ch3_consumption_lower_g_max_channel_rate_1: bool
    event_ch3_consumption_lower_g_max_channel_rate_2: bool
    event_ch3_consumption_lower_g_max_channel_rate_3: bool
    event_ch3_no_coolant_channel_rate_1: bool
    event_ch3_no_coolant_channel_rate_2: bool
    event_ch3_no_coolant_channel_rate_3: bool
    event_ch3_circuit_break_stimulation_channel_rate_1: bool
    event_ch3_circuit_break_stimulation_channel_rate_2: bool
    event_ch3_circuit_break_pressure_sensor_1: bool
    event_ch3_circuit_break_pressure_sensor_2: bool
    event_ch3_circuit_break_pressure_sensor_3: bool
    event_ch3_reverse: bool
    event_ch4_circuit_break_t_sensor_1: bool
    event_ch4_circuit_break_t_sensor_2: bool
    event_ch4_circuit_break_t_sensor_3: bool
    event_ch4_error_delta_t: bool
    event_ch4_consumption_lower_g_min_channel_rate_1: bool
    event_ch4_consumption_lower_g_min_channel_rate_2: bool
    event_ch4_consumption_lower_g_min_channel_rate_3: bool
    event_ch4_consumption_lower_g_max_channel_rate_1: bool
    event_ch4_consumption_lower_g_max_channel_rate_2: bool
    event_ch4_consumption_lower_g_max_channel_rate_3: bool
    event_ch4_no_coolant_channel_rate_1: bool
    event_ch4_no_coolant_channel_rate_2: bool
    event_ch4_no_coolant_channel_rate_3: bool
    event_ch4_circuit_break_stimulation_channel_rate_1: bool
    event_ch4_circuit_break_stimulation_channel_rate_2: bool
    event_ch4_circuit_break_pressure_sensor_1: bool
    event_ch4_circuit_break_pressure_sensor_2: bool
    event_ch4_circuit_break_pressure_sensor_3: bool
    event_ch4_reverse: bool

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((5) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((0) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((0) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((1) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        isinstance(data.days_ago, (int, timedelta))
        days_ago_tmp1 = int(data.days_ago.total_seconds() // 86400 if isinstance(data.days_ago, timedelta) else data.days_ago // 86400)
        assert 0 <= days_ago_tmp1 <= 63
        result |= ((days_ago_tmp1) & (2 ** (6) - 1)) << size
        size += 6
        assert isinstance(data.value_ch1, (int, float))
        result |= ((int(round(float(data.value_ch1) * 1000.0, 0)) & 1099511627775) & (2 ** (40) - 1)) << size
        size += 40
        assert isinstance(data.value_ch2, (int, float))
        result |= ((int(round(float(data.value_ch2) * 1000.0, 0)) & 1099511627775) & (2 ** (40) - 1)) << size
        size += 40
        assert isinstance(data.value_ch3, (int, float))
        result |= ((int(round(float(data.value_ch3) * 1000.0, 0)) & 1099511627775) & (2 ** (40) - 1)) << size
        size += 40
        assert isinstance(data.value_ch4, (int, float))
        result |= ((int(round(float(data.value_ch4) * 1000.0, 0)) & 1099511627775) & (2 ** (40) - 1)) << size
        size += 40
        isinstance(data.uptime_ch1, (int, timedelta))
        value_int_tmp2 = int(data.uptime_ch1.total_seconds() // 60 if isinstance(data.uptime_ch1, timedelta) else data.uptime_ch1 // 60) & 4194303
        result |= ((value_int_tmp2) & (2 ** (22) - 1)) << size
        size += 22
        isinstance(data.uptime_ch2, (int, timedelta))
        value_int_tmp3 = int(data.uptime_ch2.total_seconds() // 60 if isinstance(data.uptime_ch2, timedelta) else data.uptime_ch2 // 60) & 4194303
        result |= ((value_int_tmp3) & (2 ** (22) - 1)) << size
        size += 22
        isinstance(data.uptime_ch3, (int, timedelta))
        value_int_tmp4 = int(data.uptime_ch3.total_seconds() // 60 if isinstance(data.uptime_ch3, timedelta) else data.uptime_ch3 // 60) & 4194303
        result |= ((value_int_tmp4) & (2 ** (22) - 1)) << size
        size += 22
        isinstance(data.uptime_ch4, (int, timedelta))
        value_int_tmp5 = int(data.uptime_ch4.total_seconds() // 60 if isinstance(data.uptime_ch4, timedelta) else data.uptime_ch4 // 60) & 4194303
        result |= ((value_int_tmp5) & (2 ** (22) - 1)) << size
        size += 22
        assert isinstance(data.event_device_power_supply_lost, bool)
        result |= ((int(data.event_device_power_supply_lost)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_power_supply_enabled, bool)
        result |= ((int(data.event_device_power_supply_enabled)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_common_settings_changed, bool)
        result |= ((int(data.event_device_common_settings_changed)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_measure_channels_settings_changed, bool)
        result |= ((int(data.event_device_measure_channels_settings_changed)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_digital_i_o_settings_changed, bool)
        result |= ((int(data.event_device_digital_i_o_settings_changed)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_datetime_changed, bool)
        result |= ((int(data.event_device_datetime_changed)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_ethernet_settings_changed, bool)
        result |= ((int(data.event_device_ethernet_settings_changed)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_system_settings_changed1, bool)
        result |= ((int(data.event_device_system_settings_changed1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_system_settings_changed2, bool)
        result |= ((int(data.event_device_system_settings_changed2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_system_settings_changed3, bool)
        result |= ((int(data.event_device_system_settings_changed3)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_system_settings_changed4, bool)
        result |= ((int(data.event_device_system_settings_changed4)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_alarm_digital_input1, bool)
        result |= ((int(data.event_device_alarm_digital_input1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_alarm_overrate_digital_input1, bool)
        result |= ((int(data.event_device_alarm_overrate_digital_input1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_alarm_overrate_digital_input2, bool)
        result |= ((int(data.event_device_alarm_overrate_digital_input2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_alarm_underrate_digital_input1, bool)
        result |= ((int(data.event_device_alarm_underrate_digital_input1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_alarm_underrate_digital_input2, bool)
        result |= ((int(data.event_device_alarm_underrate_digital_input2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_alarm_t_high_digital_input1, bool)
        result |= ((int(data.event_device_alarm_t_high_digital_input1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_alarm_t_high_digital_input2, bool)
        result |= ((int(data.event_device_alarm_t_high_digital_input2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_alarm_t_low_digital_input1, bool)
        result |= ((int(data.event_device_alarm_t_low_digital_input1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_alarm_t_low_digital_input2, bool)
        result |= ((int(data.event_device_alarm_t_low_digital_input2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_alarm_delta_t_high_digital_input1, bool)
        result |= ((int(data.event_device_alarm_delta_t_high_digital_input1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_alarm_delta_t_high_digital_input2, bool)
        result |= ((int(data.event_device_alarm_delta_t_high_digital_input2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_alarm_delta_t_low_digital_input1, bool)
        result |= ((int(data.event_device_alarm_delta_t_low_digital_input1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_alarm_delta_t_low_digital_input2, bool)
        result |= ((int(data.event_device_alarm_delta_t_low_digital_input2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_alarm_power_high_digital_input1, bool)
        result |= ((int(data.event_device_alarm_power_high_digital_input1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_alarm_power_high_digital_input2, bool)
        result |= ((int(data.event_device_alarm_power_high_digital_input2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_alarm_power_low_digital_input1, bool)
        result |= ((int(data.event_device_alarm_power_low_digital_input1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_device_alarm_power_low_digital_input2, bool)
        result |= ((int(data.event_device_alarm_power_low_digital_input2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch1_circuit_break_t_sensor_1, bool)
        result |= ((int(data.event_ch1_circuit_break_t_sensor_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch1_circuit_break_t_sensor_2, bool)
        result |= ((int(data.event_ch1_circuit_break_t_sensor_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch1_circuit_break_t_sensor_3, bool)
        result |= ((int(data.event_ch1_circuit_break_t_sensor_3)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch1_error_delta_t, bool)
        result |= ((int(data.event_ch1_error_delta_t)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch1_consumption_lower_g_min_channel_rate_1, bool)
        result |= ((int(data.event_ch1_consumption_lower_g_min_channel_rate_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch1_consumption_lower_g_min_channel_rate_2, bool)
        result |= ((int(data.event_ch1_consumption_lower_g_min_channel_rate_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch1_consumption_lower_g_min_channel_rate_3, bool)
        result |= ((int(data.event_ch1_consumption_lower_g_min_channel_rate_3)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch1_consumption_lower_g_max_channel_rate_1, bool)
        result |= ((int(data.event_ch1_consumption_lower_g_max_channel_rate_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch1_consumption_lower_g_max_channel_rate_2, bool)
        result |= ((int(data.event_ch1_consumption_lower_g_max_channel_rate_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch1_consumption_lower_g_max_channel_rate_3, bool)
        result |= ((int(data.event_ch1_consumption_lower_g_max_channel_rate_3)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch1_no_coolant_channel_rate_1, bool)
        result |= ((int(data.event_ch1_no_coolant_channel_rate_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch1_no_coolant_channel_rate_2, bool)
        result |= ((int(data.event_ch1_no_coolant_channel_rate_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch1_no_coolant_channel_rate_3, bool)
        result |= ((int(data.event_ch1_no_coolant_channel_rate_3)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch1_circuit_break_stimulation_channel_rate_1, bool)
        result |= ((int(data.event_ch1_circuit_break_stimulation_channel_rate_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch1_circuit_break_stimulation_channel_rate_2, bool)
        result |= ((int(data.event_ch1_circuit_break_stimulation_channel_rate_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch1_circuit_break_pressure_sensor_1, bool)
        result |= ((int(data.event_ch1_circuit_break_pressure_sensor_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch1_circuit_break_pressure_sensor_2, bool)
        result |= ((int(data.event_ch1_circuit_break_pressure_sensor_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch1_circuit_break_pressure_sensor_3, bool)
        result |= ((int(data.event_ch1_circuit_break_pressure_sensor_3)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch1_reverse, bool)
        result |= ((int(data.event_ch1_reverse)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch2_circuit_break_t_sensor_1, bool)
        result |= ((int(data.event_ch2_circuit_break_t_sensor_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch2_circuit_break_t_sensor_2, bool)
        result |= ((int(data.event_ch2_circuit_break_t_sensor_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch2_circuit_break_t_sensor_3, bool)
        result |= ((int(data.event_ch2_circuit_break_t_sensor_3)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch2_error_delta_t, bool)
        result |= ((int(data.event_ch2_error_delta_t)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch2_consumption_lower_g_min_channel_rate_1, bool)
        result |= ((int(data.event_ch2_consumption_lower_g_min_channel_rate_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch2_consumption_lower_g_min_channel_rate_2, bool)
        result |= ((int(data.event_ch2_consumption_lower_g_min_channel_rate_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch2_consumption_lower_g_min_channel_rate_3, bool)
        result |= ((int(data.event_ch2_consumption_lower_g_min_channel_rate_3)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch2_consumption_lower_g_max_channel_rate_1, bool)
        result |= ((int(data.event_ch2_consumption_lower_g_max_channel_rate_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch2_consumption_lower_g_max_channel_rate_2, bool)
        result |= ((int(data.event_ch2_consumption_lower_g_max_channel_rate_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch2_consumption_lower_g_max_channel_rate_3, bool)
        result |= ((int(data.event_ch2_consumption_lower_g_max_channel_rate_3)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch2_no_coolant_channel_rate_1, bool)
        result |= ((int(data.event_ch2_no_coolant_channel_rate_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch2_no_coolant_channel_rate_2, bool)
        result |= ((int(data.event_ch2_no_coolant_channel_rate_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch2_no_coolant_channel_rate_3, bool)
        result |= ((int(data.event_ch2_no_coolant_channel_rate_3)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch2_circuit_break_stimulation_channel_rate_1, bool)
        result |= ((int(data.event_ch2_circuit_break_stimulation_channel_rate_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch2_circuit_break_stimulation_channel_rate_2, bool)
        result |= ((int(data.event_ch2_circuit_break_stimulation_channel_rate_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch2_circuit_break_pressure_sensor_1, bool)
        result |= ((int(data.event_ch2_circuit_break_pressure_sensor_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch2_circuit_break_pressure_sensor_2, bool)
        result |= ((int(data.event_ch2_circuit_break_pressure_sensor_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch2_circuit_break_pressure_sensor_3, bool)
        result |= ((int(data.event_ch2_circuit_break_pressure_sensor_3)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch2_reverse, bool)
        result |= ((int(data.event_ch2_reverse)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch3_circuit_break_t_sensor_1, bool)
        result |= ((int(data.event_ch3_circuit_break_t_sensor_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch3_circuit_break_t_sensor_2, bool)
        result |= ((int(data.event_ch3_circuit_break_t_sensor_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch3_circuit_break_t_sensor_3, bool)
        result |= ((int(data.event_ch3_circuit_break_t_sensor_3)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch3_error_delta_t, bool)
        result |= ((int(data.event_ch3_error_delta_t)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch3_consumption_lower_g_min_channel_rate_1, bool)
        result |= ((int(data.event_ch3_consumption_lower_g_min_channel_rate_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch3_consumption_lower_g_min_channel_rate_2, bool)
        result |= ((int(data.event_ch3_consumption_lower_g_min_channel_rate_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch3_consumption_lower_g_min_channel_rate_3, bool)
        result |= ((int(data.event_ch3_consumption_lower_g_min_channel_rate_3)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch3_consumption_lower_g_max_channel_rate_1, bool)
        result |= ((int(data.event_ch3_consumption_lower_g_max_channel_rate_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch3_consumption_lower_g_max_channel_rate_2, bool)
        result |= ((int(data.event_ch3_consumption_lower_g_max_channel_rate_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch3_consumption_lower_g_max_channel_rate_3, bool)
        result |= ((int(data.event_ch3_consumption_lower_g_max_channel_rate_3)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch3_no_coolant_channel_rate_1, bool)
        result |= ((int(data.event_ch3_no_coolant_channel_rate_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch3_no_coolant_channel_rate_2, bool)
        result |= ((int(data.event_ch3_no_coolant_channel_rate_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch3_no_coolant_channel_rate_3, bool)
        result |= ((int(data.event_ch3_no_coolant_channel_rate_3)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch3_circuit_break_stimulation_channel_rate_1, bool)
        result |= ((int(data.event_ch3_circuit_break_stimulation_channel_rate_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch3_circuit_break_stimulation_channel_rate_2, bool)
        result |= ((int(data.event_ch3_circuit_break_stimulation_channel_rate_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch3_circuit_break_pressure_sensor_1, bool)
        result |= ((int(data.event_ch3_circuit_break_pressure_sensor_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch3_circuit_break_pressure_sensor_2, bool)
        result |= ((int(data.event_ch3_circuit_break_pressure_sensor_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch3_circuit_break_pressure_sensor_3, bool)
        result |= ((int(data.event_ch3_circuit_break_pressure_sensor_3)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch3_reverse, bool)
        result |= ((int(data.event_ch3_reverse)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch4_circuit_break_t_sensor_1, bool)
        result |= ((int(data.event_ch4_circuit_break_t_sensor_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch4_circuit_break_t_sensor_2, bool)
        result |= ((int(data.event_ch4_circuit_break_t_sensor_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch4_circuit_break_t_sensor_3, bool)
        result |= ((int(data.event_ch4_circuit_break_t_sensor_3)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch4_error_delta_t, bool)
        result |= ((int(data.event_ch4_error_delta_t)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch4_consumption_lower_g_min_channel_rate_1, bool)
        result |= ((int(data.event_ch4_consumption_lower_g_min_channel_rate_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch4_consumption_lower_g_min_channel_rate_2, bool)
        result |= ((int(data.event_ch4_consumption_lower_g_min_channel_rate_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch4_consumption_lower_g_min_channel_rate_3, bool)
        result |= ((int(data.event_ch4_consumption_lower_g_min_channel_rate_3)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch4_consumption_lower_g_max_channel_rate_1, bool)
        result |= ((int(data.event_ch4_consumption_lower_g_max_channel_rate_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch4_consumption_lower_g_max_channel_rate_2, bool)
        result |= ((int(data.event_ch4_consumption_lower_g_max_channel_rate_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch4_consumption_lower_g_max_channel_rate_3, bool)
        result |= ((int(data.event_ch4_consumption_lower_g_max_channel_rate_3)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch4_no_coolant_channel_rate_1, bool)
        result |= ((int(data.event_ch4_no_coolant_channel_rate_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch4_no_coolant_channel_rate_2, bool)
        result |= ((int(data.event_ch4_no_coolant_channel_rate_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch4_no_coolant_channel_rate_3, bool)
        result |= ((int(data.event_ch4_no_coolant_channel_rate_3)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch4_circuit_break_stimulation_channel_rate_1, bool)
        result |= ((int(data.event_ch4_circuit_break_stimulation_channel_rate_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch4_circuit_break_stimulation_channel_rate_2, bool)
        result |= ((int(data.event_ch4_circuit_break_stimulation_channel_rate_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch4_circuit_break_pressure_sensor_1, bool)
        result |= ((int(data.event_ch4_circuit_break_pressure_sensor_1)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch4_circuit_break_pressure_sensor_2, bool)
        result |= ((int(data.event_ch4_circuit_break_pressure_sensor_2)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch4_circuit_break_pressure_sensor_3, bool)
        result |= ((int(data.event_ch4_circuit_break_pressure_sensor_3)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_ch4_reverse, bool)
        result |= ((int(data.event_ch4_reverse)) & (2 ** (1) - 1)) << size
        size += 1
        return result.to_bytes(48, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceHeatGroupMeterDailyData':
        result__el_tmp6: Dict[str, Any] = dict()
        if 5 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp6["days_ago"] = timedelta(seconds=buf.shift(6) * 86400)
        result__el_tmp6["value_ch1"] = round(buf.shift(40) / 1000.0, 3)
        result__el_tmp6["value_ch2"] = round(buf.shift(40) / 1000.0, 3)
        result__el_tmp6["value_ch3"] = round(buf.shift(40) / 1000.0, 3)
        result__el_tmp6["value_ch4"] = round(buf.shift(40) / 1000.0, 3)
        result__el_tmp6["uptime_ch1"] = timedelta(seconds=buf.shift(22) * 60)
        result__el_tmp6["uptime_ch2"] = timedelta(seconds=buf.shift(22) * 60)
        result__el_tmp6["uptime_ch3"] = timedelta(seconds=buf.shift(22) * 60)
        result__el_tmp6["uptime_ch4"] = timedelta(seconds=buf.shift(22) * 60)
        result__el_tmp6["event_device_power_supply_lost"] = bool(buf.shift(1))
        result__el_tmp6["event_device_power_supply_enabled"] = bool(buf.shift(1))
        result__el_tmp6["event_device_common_settings_changed"] = bool(buf.shift(1))
        result__el_tmp6["event_device_measure_channels_settings_changed"] = bool(buf.shift(1))
        result__el_tmp6["event_device_digital_i_o_settings_changed"] = bool(buf.shift(1))
        result__el_tmp6["event_device_datetime_changed"] = bool(buf.shift(1))
        result__el_tmp6["event_device_ethernet_settings_changed"] = bool(buf.shift(1))
        result__el_tmp6["event_device_system_settings_changed1"] = bool(buf.shift(1))
        result__el_tmp6["event_device_system_settings_changed2"] = bool(buf.shift(1))
        result__el_tmp6["event_device_system_settings_changed3"] = bool(buf.shift(1))
        result__el_tmp6["event_device_system_settings_changed4"] = bool(buf.shift(1))
        result__el_tmp6["event_device_alarm_digital_input1"] = bool(buf.shift(1))
        result__el_tmp6["event_device_alarm_overrate_digital_input1"] = bool(buf.shift(1))
        result__el_tmp6["event_device_alarm_overrate_digital_input2"] = bool(buf.shift(1))
        result__el_tmp6["event_device_alarm_underrate_digital_input1"] = bool(buf.shift(1))
        result__el_tmp6["event_device_alarm_underrate_digital_input2"] = bool(buf.shift(1))
        result__el_tmp6["event_device_alarm_t_high_digital_input1"] = bool(buf.shift(1))
        result__el_tmp6["event_device_alarm_t_high_digital_input2"] = bool(buf.shift(1))
        result__el_tmp6["event_device_alarm_t_low_digital_input1"] = bool(buf.shift(1))
        result__el_tmp6["event_device_alarm_t_low_digital_input2"] = bool(buf.shift(1))
        result__el_tmp6["event_device_alarm_delta_t_high_digital_input1"] = bool(buf.shift(1))
        result__el_tmp6["event_device_alarm_delta_t_high_digital_input2"] = bool(buf.shift(1))
        result__el_tmp6["event_device_alarm_delta_t_low_digital_input1"] = bool(buf.shift(1))
        result__el_tmp6["event_device_alarm_delta_t_low_digital_input2"] = bool(buf.shift(1))
        result__el_tmp6["event_device_alarm_power_high_digital_input1"] = bool(buf.shift(1))
        result__el_tmp6["event_device_alarm_power_high_digital_input2"] = bool(buf.shift(1))
        result__el_tmp6["event_device_alarm_power_low_digital_input1"] = bool(buf.shift(1))
        result__el_tmp6["event_device_alarm_power_low_digital_input2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch1_circuit_break_t_sensor_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch1_circuit_break_t_sensor_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch1_circuit_break_t_sensor_3"] = bool(buf.shift(1))
        result__el_tmp6["event_ch1_error_delta_t"] = bool(buf.shift(1))
        result__el_tmp6["event_ch1_consumption_lower_g_min_channel_rate_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch1_consumption_lower_g_min_channel_rate_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch1_consumption_lower_g_min_channel_rate_3"] = bool(buf.shift(1))
        result__el_tmp6["event_ch1_consumption_lower_g_max_channel_rate_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch1_consumption_lower_g_max_channel_rate_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch1_consumption_lower_g_max_channel_rate_3"] = bool(buf.shift(1))
        result__el_tmp6["event_ch1_no_coolant_channel_rate_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch1_no_coolant_channel_rate_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch1_no_coolant_channel_rate_3"] = bool(buf.shift(1))
        result__el_tmp6["event_ch1_circuit_break_stimulation_channel_rate_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch1_circuit_break_stimulation_channel_rate_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch1_circuit_break_pressure_sensor_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch1_circuit_break_pressure_sensor_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch1_circuit_break_pressure_sensor_3"] = bool(buf.shift(1))
        result__el_tmp6["event_ch1_reverse"] = bool(buf.shift(1))
        result__el_tmp6["event_ch2_circuit_break_t_sensor_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch2_circuit_break_t_sensor_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch2_circuit_break_t_sensor_3"] = bool(buf.shift(1))
        result__el_tmp6["event_ch2_error_delta_t"] = bool(buf.shift(1))
        result__el_tmp6["event_ch2_consumption_lower_g_min_channel_rate_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch2_consumption_lower_g_min_channel_rate_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch2_consumption_lower_g_min_channel_rate_3"] = bool(buf.shift(1))
        result__el_tmp6["event_ch2_consumption_lower_g_max_channel_rate_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch2_consumption_lower_g_max_channel_rate_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch2_consumption_lower_g_max_channel_rate_3"] = bool(buf.shift(1))
        result__el_tmp6["event_ch2_no_coolant_channel_rate_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch2_no_coolant_channel_rate_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch2_no_coolant_channel_rate_3"] = bool(buf.shift(1))
        result__el_tmp6["event_ch2_circuit_break_stimulation_channel_rate_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch2_circuit_break_stimulation_channel_rate_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch2_circuit_break_pressure_sensor_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch2_circuit_break_pressure_sensor_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch2_circuit_break_pressure_sensor_3"] = bool(buf.shift(1))
        result__el_tmp6["event_ch2_reverse"] = bool(buf.shift(1))
        result__el_tmp6["event_ch3_circuit_break_t_sensor_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch3_circuit_break_t_sensor_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch3_circuit_break_t_sensor_3"] = bool(buf.shift(1))
        result__el_tmp6["event_ch3_error_delta_t"] = bool(buf.shift(1))
        result__el_tmp6["event_ch3_consumption_lower_g_min_channel_rate_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch3_consumption_lower_g_min_channel_rate_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch3_consumption_lower_g_min_channel_rate_3"] = bool(buf.shift(1))
        result__el_tmp6["event_ch3_consumption_lower_g_max_channel_rate_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch3_consumption_lower_g_max_channel_rate_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch3_consumption_lower_g_max_channel_rate_3"] = bool(buf.shift(1))
        result__el_tmp6["event_ch3_no_coolant_channel_rate_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch3_no_coolant_channel_rate_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch3_no_coolant_channel_rate_3"] = bool(buf.shift(1))
        result__el_tmp6["event_ch3_circuit_break_stimulation_channel_rate_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch3_circuit_break_stimulation_channel_rate_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch3_circuit_break_pressure_sensor_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch3_circuit_break_pressure_sensor_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch3_circuit_break_pressure_sensor_3"] = bool(buf.shift(1))
        result__el_tmp6["event_ch3_reverse"] = bool(buf.shift(1))
        result__el_tmp6["event_ch4_circuit_break_t_sensor_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch4_circuit_break_t_sensor_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch4_circuit_break_t_sensor_3"] = bool(buf.shift(1))
        result__el_tmp6["event_ch4_error_delta_t"] = bool(buf.shift(1))
        result__el_tmp6["event_ch4_consumption_lower_g_min_channel_rate_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch4_consumption_lower_g_min_channel_rate_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch4_consumption_lower_g_min_channel_rate_3"] = bool(buf.shift(1))
        result__el_tmp6["event_ch4_consumption_lower_g_max_channel_rate_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch4_consumption_lower_g_max_channel_rate_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch4_consumption_lower_g_max_channel_rate_3"] = bool(buf.shift(1))
        result__el_tmp6["event_ch4_no_coolant_channel_rate_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch4_no_coolant_channel_rate_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch4_no_coolant_channel_rate_3"] = bool(buf.shift(1))
        result__el_tmp6["event_ch4_circuit_break_stimulation_channel_rate_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch4_circuit_break_stimulation_channel_rate_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch4_circuit_break_pressure_sensor_1"] = bool(buf.shift(1))
        result__el_tmp6["event_ch4_circuit_break_pressure_sensor_2"] = bool(buf.shift(1))
        result__el_tmp6["event_ch4_circuit_break_pressure_sensor_3"] = bool(buf.shift(1))
        result__el_tmp6["event_ch4_reverse"] = bool(buf.shift(1))
        result = SmpmUlDeviceHeatGroupMeterDailyData(**result__el_tmp6)
        buf.shift(9)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        overloading_value = Decimal(str(134217.727))
        return [
            IntegrationV0MessageData(
                dt=days_ago_calculation(received_at, device_tz, time(0), self.days_ago),
                consumption=[
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        value=Decimal(str(self.value_ch1)),
                        resource_type=ResourceType.COMMON,
                        channel=1,
                        overloading_value=overloading_value,
                        journal_data_type=JournalDataType.END_OF_DAY if self.days_ago else JournalDataType.CURRENT,
                    ),
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        value=Decimal(str(self.value_ch2)),
                        resource_type=ResourceType.COMMON,
                        channel=2,
                        overloading_value=overloading_value,
                        journal_data_type=JournalDataType.END_OF_DAY if self.days_ago else JournalDataType.CURRENT,
                    ),
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        value=Decimal(str(self.value_ch3)),
                        resource_type=ResourceType.COMMON,
                        channel=3,
                        overloading_value=overloading_value,
                        journal_data_type=JournalDataType.END_OF_DAY if self.days_ago else JournalDataType.CURRENT,
                    ),
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        value=Decimal(str(self.value_ch4)),
                        resource_type=ResourceType.COMMON,
                        channel=4,
                        overloading_value=overloading_value,
                        journal_data_type=JournalDataType.END_OF_DAY if self.days_ago else JournalDataType.CURRENT,
                    ),
                ],
                events=[
                    *([IntegrationV0MessageEvent.DEVICE__POWER_SUPPLY_LOST] if self.event_device_power_supply_lost else []),
                    *([IntegrationV0MessageEvent.DEVICE__POWER_SUPPLY_ENABLED] if self.event_device_power_supply_enabled else []),
                    *([IntegrationV0MessageEvent.DEVICE__COMMON_SETTINGS_CHANGED] if self.event_device_common_settings_changed else []),
                    *([IntegrationV0MessageEvent.DEVICE__MEASURE_CHANNELS_SETTINGS_CHANGED] if self.event_device_measure_channels_settings_changed else []),
                    *([IntegrationV0MessageEvent.DEVICE__DIGITAL_I_O_SETTINGS_CHANGED] if self.event_device_digital_i_o_settings_changed else []),
                    *([IntegrationV0MessageEvent.DEVICE__DATETIME_CHANGED] if self.event_device_datetime_changed else []),
                    *([IntegrationV0MessageEvent.DEVICE__ETHERNET_SETTINGS_CHANGED] if self.event_device_ethernet_settings_changed else []),
                    *([IntegrationV0MessageEvent.DEVICE__SYSTEM_SETTINGS_CHANGED_1] if self.event_device_system_settings_changed1 else []),
                    *([IntegrationV0MessageEvent.DEVICE__SYSTEM_SETTINGS_CHANGED_2] if self.event_device_system_settings_changed2 else []),
                    *([IntegrationV0MessageEvent.DEVICE__SYSTEM_SETTINGS_CHANGED_3] if self.event_device_system_settings_changed3 else []),
                    *([IntegrationV0MessageEvent.DEVICE__SYSTEM_SETTINGS_CHANGED_4] if self.event_device_system_settings_changed4 else []),
                    *([IntegrationV0MessageEvent.DEVICE__ALARM_DIGITAL_INPUT_1] if self.event_device_alarm_digital_input1 else []),
                    *([IntegrationV0MessageEvent.DEVICE__ALARM_OVERRATE_DIGITAL_INPUT_1] if self.event_device_alarm_overrate_digital_input1 else []),
                    *([IntegrationV0MessageEvent.DEVICE__ALARM_OVERRATE_DIGITAL_INPUT_2] if self.event_device_alarm_overrate_digital_input2 else []),
                    *([IntegrationV0MessageEvent.DEVICE__ALARM_UNDERRATE_DIGITAL_INPUT_1] if self.event_device_alarm_underrate_digital_input1 else []),
                    *([IntegrationV0MessageEvent.DEVICE__ALARM_UNDERRATE_DIGITAL_INPUT_2] if self.event_device_alarm_underrate_digital_input2 else []),
                    *([IntegrationV0MessageEvent.DEVICE__ALARM_T_HIGH_DIGITAL_INPUT_1] if self.event_device_alarm_t_high_digital_input1 else []),
                    *([IntegrationV0MessageEvent.DEVICE__ALARM_T_HIGH_DIGITAL_INPUT_2] if self.event_device_alarm_t_high_digital_input2 else []),
                    *([IntegrationV0MessageEvent.DEVICE__ALARM_T_LOW_DIGITAL_INPUT_1] if self.event_device_alarm_t_low_digital_input1 else []),
                    *([IntegrationV0MessageEvent.DEVICE__ALARM_T_LOW_DIGITAL_INPUT_2] if self.event_device_alarm_t_low_digital_input2 else []),
                    *([IntegrationV0MessageEvent.DEVICE__ALARM_DELTA_T_HIGH_DIGITAL_INPUT_1] if self.event_device_alarm_delta_t_high_digital_input1 else []),
                    *([IntegrationV0MessageEvent.DEVICE__ALARM_DELTA_T_HIGH_DIGITAL_INPUT_2] if self.event_device_alarm_delta_t_high_digital_input2 else []),
                    *([IntegrationV0MessageEvent.DEVICE__ALARM_DELTA_T_LOW_DIGITAL_INPUT_1] if self.event_device_alarm_delta_t_low_digital_input1 else []),
                    *([IntegrationV0MessageEvent.DEVICE__ALARM_DELTA_T_LOW_DIGITAL_INPUT_2] if self.event_device_alarm_delta_t_low_digital_input2 else []),
                    *([IntegrationV0MessageEvent.DEVICE__ALARM_POWER_HIGH_DIGITAL_INPUT_1] if self.event_device_alarm_power_high_digital_input1 else []),
                    *([IntegrationV0MessageEvent.DEVICE__ALARM_POWER_HIGH_DIGITAL_INPUT_2] if self.event_device_alarm_power_high_digital_input2 else []),
                    *([IntegrationV0MessageEvent.DEVICE__ALARM_POWER_LOW_DIGITAL_INPUT_1] if self.event_device_alarm_power_low_digital_input1 else []),
                    *([IntegrationV0MessageEvent.DEVICE__ALARM_POWER_LOW_DIGITAL_INPUT_2] if self.event_device_alarm_power_low_digital_input2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_T_SENSOR_1] if self.event_ch1_circuit_break_t_sensor_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_T_SENSOR_2] if self.event_ch1_circuit_break_t_sensor_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_T_SENSOR_3] if self.event_ch1_circuit_break_t_sensor_3 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__ERROR_DELTA_T] if self.event_ch1_error_delta_t else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MIN_CHANNEL_RATE_1] if self.event_ch1_consumption_lower_g_min_channel_rate_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MIN_CHANNEL_RATE_2] if self.event_ch1_consumption_lower_g_min_channel_rate_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MIN_CHANNEL_RATE_3] if self.event_ch1_consumption_lower_g_min_channel_rate_3 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MAX_CHANNEL_RATE_1] if self.event_ch1_consumption_lower_g_max_channel_rate_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MAX_CHANNEL_RATE_2] if self.event_ch1_consumption_lower_g_max_channel_rate_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MAX_CHANNEL_RATE_3] if self.event_ch1_consumption_lower_g_max_channel_rate_3 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__NO_COOLANT_CHANNEL_RATE_1] if self.event_ch1_no_coolant_channel_rate_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__NO_COOLANT_CHANNEL_RATE_2] if self.event_ch1_no_coolant_channel_rate_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__NO_COOLANT_CHANNEL_RATE_3] if self.event_ch1_no_coolant_channel_rate_3 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_STIMULATION_CHANNEL_RATE_1] if self.event_ch1_circuit_break_stimulation_channel_rate_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_STIMULATION_CHANNEL_RATE_2] if self.event_ch1_circuit_break_stimulation_channel_rate_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_PRESSURE_SENSOR_1] if self.event_ch1_circuit_break_pressure_sensor_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_PRESSURE_SENSOR_2] if self.event_ch1_circuit_break_pressure_sensor_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_PRESSURE_SENSOR_3] if self.event_ch1_circuit_break_pressure_sensor_3 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__REVERSE] if self.event_ch1_reverse else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_T_SENSOR_1] if self.event_ch2_circuit_break_t_sensor_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_T_SENSOR_2] if self.event_ch2_circuit_break_t_sensor_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_T_SENSOR_3] if self.event_ch2_circuit_break_t_sensor_3 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__ERROR_DELTA_T] if self.event_ch2_error_delta_t else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MIN_CHANNEL_RATE_1] if self.event_ch2_consumption_lower_g_min_channel_rate_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MIN_CHANNEL_RATE_2] if self.event_ch2_consumption_lower_g_min_channel_rate_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MIN_CHANNEL_RATE_3] if self.event_ch2_consumption_lower_g_min_channel_rate_3 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MAX_CHANNEL_RATE_1] if self.event_ch2_consumption_lower_g_max_channel_rate_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MAX_CHANNEL_RATE_2] if self.event_ch2_consumption_lower_g_max_channel_rate_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MAX_CHANNEL_RATE_3] if self.event_ch2_consumption_lower_g_max_channel_rate_3 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__NO_COOLANT_CHANNEL_RATE_1] if self.event_ch2_no_coolant_channel_rate_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__NO_COOLANT_CHANNEL_RATE_2] if self.event_ch2_no_coolant_channel_rate_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__NO_COOLANT_CHANNEL_RATE_3] if self.event_ch2_no_coolant_channel_rate_3 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_STIMULATION_CHANNEL_RATE_1] if self.event_ch2_circuit_break_stimulation_channel_rate_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_STIMULATION_CHANNEL_RATE_2] if self.event_ch2_circuit_break_stimulation_channel_rate_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_PRESSURE_SENSOR_1] if self.event_ch2_circuit_break_pressure_sensor_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_PRESSURE_SENSOR_2] if self.event_ch2_circuit_break_pressure_sensor_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_PRESSURE_SENSOR_3] if self.event_ch2_circuit_break_pressure_sensor_3 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__REVERSE] if self.event_ch2_reverse else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_T_SENSOR_1] if self.event_ch3_circuit_break_t_sensor_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_T_SENSOR_2] if self.event_ch3_circuit_break_t_sensor_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_T_SENSOR_3] if self.event_ch3_circuit_break_t_sensor_3 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__ERROR_DELTA_T] if self.event_ch3_error_delta_t else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MIN_CHANNEL_RATE_1] if self.event_ch3_consumption_lower_g_min_channel_rate_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MIN_CHANNEL_RATE_2] if self.event_ch3_consumption_lower_g_min_channel_rate_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MIN_CHANNEL_RATE_3] if self.event_ch3_consumption_lower_g_min_channel_rate_3 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MAX_CHANNEL_RATE_1] if self.event_ch3_consumption_lower_g_max_channel_rate_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MAX_CHANNEL_RATE_2] if self.event_ch3_consumption_lower_g_max_channel_rate_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MAX_CHANNEL_RATE_3] if self.event_ch3_consumption_lower_g_max_channel_rate_3 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__NO_COOLANT_CHANNEL_RATE_1] if self.event_ch3_no_coolant_channel_rate_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__NO_COOLANT_CHANNEL_RATE_2] if self.event_ch3_no_coolant_channel_rate_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__NO_COOLANT_CHANNEL_RATE_3] if self.event_ch3_no_coolant_channel_rate_3 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_STIMULATION_CHANNEL_RATE_1] if self.event_ch3_circuit_break_stimulation_channel_rate_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_STIMULATION_CHANNEL_RATE_2] if self.event_ch3_circuit_break_stimulation_channel_rate_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_PRESSURE_SENSOR_1] if self.event_ch3_circuit_break_pressure_sensor_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_PRESSURE_SENSOR_2] if self.event_ch3_circuit_break_pressure_sensor_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_PRESSURE_SENSOR_3] if self.event_ch3_circuit_break_pressure_sensor_3 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__REVERSE] if self.event_ch3_reverse else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_T_SENSOR_1] if self.event_ch4_circuit_break_t_sensor_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_T_SENSOR_2] if self.event_ch4_circuit_break_t_sensor_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_T_SENSOR_3] if self.event_ch4_circuit_break_t_sensor_3 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__ERROR_DELTA_T] if self.event_ch4_error_delta_t else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MIN_CHANNEL_RATE_1] if self.event_ch4_consumption_lower_g_min_channel_rate_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MIN_CHANNEL_RATE_2] if self.event_ch4_consumption_lower_g_min_channel_rate_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MIN_CHANNEL_RATE_3] if self.event_ch4_consumption_lower_g_min_channel_rate_3 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MAX_CHANNEL_RATE_1] if self.event_ch4_consumption_lower_g_max_channel_rate_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MAX_CHANNEL_RATE_2] if self.event_ch4_consumption_lower_g_max_channel_rate_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CONSUMPTION_LOWER_G_MAX_CHANNEL_RATE_3] if self.event_ch4_consumption_lower_g_max_channel_rate_3 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__NO_COOLANT_CHANNEL_RATE_1] if self.event_ch4_no_coolant_channel_rate_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__NO_COOLANT_CHANNEL_RATE_2] if self.event_ch4_no_coolant_channel_rate_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__NO_COOLANT_CHANNEL_RATE_3] if self.event_ch4_no_coolant_channel_rate_3 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_STIMULATION_CHANNEL_RATE_1] if self.event_ch4_circuit_break_stimulation_channel_rate_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_STIMULATION_CHANNEL_RATE_2] if self.event_ch4_circuit_break_stimulation_channel_rate_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_PRESSURE_SENSOR_1] if self.event_ch4_circuit_break_pressure_sensor_1 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_PRESSURE_SENSOR_2] if self.event_ch4_circuit_break_pressure_sensor_2 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__CIRCUIT_BREAK_PRESSURE_SENSOR_3] if self.event_ch4_circuit_break_pressure_sensor_3 else []),
                    *([IntegrationV0MessageEvent.SYSTEM__REVERSE] if self.event_ch4_reverse else []),
                ],
                sensors=[
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.uptime_ch1.total_seconds())),
                        sensor_type=SensorType.UPTIME,
                        channel_id=1,
                    ),
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.uptime_ch2.total_seconds())),
                        sensor_type=SensorType.UPTIME,
                        channel_id=2,
                    ),
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.uptime_ch3.total_seconds())),
                        sensor_type=SensorType.UPTIME,
                        channel_id=3,
                    ),
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.uptime_ch4.total_seconds())),
                        sensor_type=SensorType.UPTIME,
                        channel_id=4,
                    ),
                ],

            ),
        ]



from datetime import timedelta

from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_3phase_consumed import \
    SmpmUlDeviceEnergy16B3PhaseConsumedData
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_ul_device_energy_16b_3phase_consumed() -> None:
    case_serialized = bytes.fromhex("cc02300f1e00900d0160574cdb5e0105")
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=False, total=123123, phase_a=4313, phase_b=14312123, phase_c=1312123) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=False, total=123123, phase_a=4313, phase_b=14312123, phase_c=1312123))  # noqa: E501
    case_serialized = bytes.fromhex("cc020800000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, phase_a=0, phase_b=0, phase_c=0, total=0, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("cc020800000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, phase_a=0, phase_b=0, phase_c=0, total=8589934592, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("cc020800000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, phase_a=0, phase_b=0, phase_c=67108864, total=0, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("cc020800000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, phase_a=0, phase_b=0, phase_c=67108864, total=8589934592, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("cc020800000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, phase_a=0, phase_b=67108864, phase_c=0, total=0, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("cc020800000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, phase_a=0, phase_b=67108864, phase_c=0, total=8589934592, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("cc020800000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, phase_a=0, phase_b=67108864, phase_c=67108864, total=0, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("cc020800000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, phase_a=0, phase_b=67108864, phase_c=67108864, total=8589934592, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("cc020800000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, phase_a=67108864, phase_b=0, phase_c=0, total=0, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("cc020800000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, phase_a=67108864, phase_b=0, phase_c=0, total=8589934592, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("cc020800000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, phase_a=67108864, phase_b=0, phase_c=67108864, total=0, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("cc020800000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, phase_a=67108864, phase_b=0, phase_c=67108864, total=8589934592, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("cc020800000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, phase_a=67108864, phase_b=67108864, phase_c=0, total=0, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("cc020800000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, phase_a=67108864, phase_b=67108864, phase_c=0, total=8589934592, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("cc020800000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, phase_a=67108864, phase_b=67108864, phase_c=67108864, total=0, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("cc020800000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, phase_a=67108864, phase_b=67108864, phase_c=67108864, total=8589934592, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("ccf20f00000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, phase_a=0, phase_b=0, phase_c=0, total=0, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("ccf20f00000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, phase_a=0, phase_b=0, phase_c=0, total=8589934592, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("ccf20f00000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, phase_a=0, phase_b=0, phase_c=67108864, total=0, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("ccf20f00000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, phase_a=0, phase_b=0, phase_c=67108864, total=8589934592, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("ccf20f00000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, phase_a=0, phase_b=67108864, phase_c=0, total=0, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("ccf20f00000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, phase_a=0, phase_b=67108864, phase_c=0, total=8589934592, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("ccf20f00000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, phase_a=0, phase_b=67108864, phase_c=67108864, total=0, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("ccf20f00000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, phase_a=0, phase_b=67108864, phase_c=67108864, total=8589934592, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("ccf20f00000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, phase_a=67108864, phase_b=0, phase_c=0, total=0, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("ccf20f00000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, phase_a=67108864, phase_b=0, phase_c=0, total=8589934592, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("ccf20f00000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, phase_a=67108864, phase_b=0, phase_c=67108864, total=0, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("ccf20f00000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, phase_a=67108864, phase_b=0, phase_c=67108864, total=8589934592, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("ccf20f00000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, phase_a=67108864, phase_b=67108864, phase_c=0, total=0, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("ccf20f00000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, phase_a=67108864, phase_b=67108864, phase_c=0, total=8589934592, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("ccf20f00000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, phase_a=67108864, phase_b=67108864, phase_c=67108864, total=0, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("ccf20f00000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16B3PhaseConsumedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, total=0, phase_a=0, phase_b=0, phase_c=0) == SmpmUlDeviceEnergy16B3PhaseConsumedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16B3PhaseConsumedData.serialize(SmpmUlDeviceEnergy16B3PhaseConsumedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, phase_a=67108864, phase_b=67108864, phase_c=67108864, total=8589934592, valid=True))  # noqa: E501



from typing import Any, Iterable

from data_gateway_sdk.crypto_algorithms.constants.constants_aes import KEY_LENGTH_BYTES_AES256, KEY_LENGTH_BYTES_AES128, AES_BLOCK_SIZE, STRAIGHT_BOX, INVERSE_BOX
from data_gateway_sdk.crypto_algorithms.extensions.errors import AesError
from data_gateway_sdk.crypto_algorithms.extensions.type_alias import VectorArray, VectorArrayArray, Vector


def valid_aes_key(key_bytes: bytes) -> int:
    """
    Валидация ключа (проверка чтобы ключ был ровно 16 байт или 32 байта)
    """
    if len(key_bytes) == KEY_LENGTH_BYTES_AES256:
        return KEY_LENGTH_BYTES_AES256
    elif len(key_bytes) == KEY_LENGTH_BYTES_AES128:
        return KEY_LENGTH_BYTES_AES128
    else:
        raise AesError(f"invalid length KEY '{len(key_bytes)}' was given, must be '{KEY_LENGTH_BYTES_AES128}' or '{KEY_LENGTH_BYTES_AES256}'")


def add_round_key(state: VectorArray, key_schedule: VectorArrayArray, round_number: int) -> None:
    """
    Трансформация производит побитовый XOR каждого элемента из State с соответствующим элементом из RoundKey.

    RoundKey — массив такого же размера, как и State,
    который строится для каждого раунда на основе секретного ключа функцией KeyExpansion(), которую и рассмотрим далее.

    State — промежуточный результат шифрования,
    который может быть представлен как прямоугольный массив байтов имеющий 4 строки и Nb колонок.
    Каждая ячейка State содержит значение размером в 1 байт
    """
    round_key = key_schedule[round_number]
    for round_index in range(len(state)):
        state[round_index] = [state[round_index][constant] ^ round_key[round_index][constant] for constant in range(len(state[0]))]


def sub_bytes(state: VectorArray) -> None:
    """
    Преобразование представляет собой замену каждого байта из State на соответствующий ему из константной таблицы STRAIGHT_BOX.
    Значения элементов Sbox представлены в шестнадцатеричной системе счисления.
    """
    for round_index in range(len(state)):
        state[round_index] = [STRAIGHT_BOX[state[round_index][constant]] for constant in range(len(state[0]))]


def shift_rows(state: VectorArray) -> None:
    """
    Простая трансформация.
    Она выполняет циклический сдвиг влево на 1 элемент для первой строки, на 2 для второй и на 3 для третьей.
    Нулевая строка не сдвигается.

    пример:
    [00, 10, 20, 30]     [00, 10, 20, 30]
    [01, 11, 21, 31] --> [11, 21, 31, 01]
    [02, 12, 22, 32]     [22, 32, 02, 12]
    [03, 13, 23, 33]     [33, 03, 13, 23]
    """
    state[0][1], state[1][1], state[2][1], state[3][1] = state[1][1], state[2][1], state[3][1], state[0][1]
    state[0][2], state[1][2], state[2][2], state[3][2] = state[2][2], state[3][2], state[0][2], state[1][2]
    state[0][3], state[1][3], state[2][3], state[3][3] = state[3][3], state[0][3], state[1][3], state[2][3]


def xtime(value: int) -> int:
    """
    Функция, которая выполняет умножение на соответствующую константу в GF(28) по правилам.
    """
    if value & 0x80:
        return ((value << 1) ^ 0x1b) & 0xff
    return value << 1


def mix_column(column: Vector) -> None:
    """
    В рамках этой трансформации каждая колонка в State представляется в виде многочлена
    и перемножается в поле GF(2^8) по модулю x^4 + 1 с фиксированным многочленом 3x^3 + x^2 + x + 2.
    """
    column_zero = column[0]
    all_xor = column[0] ^ column[1] ^ column[2] ^ column[3]
    column[0] ^= all_xor ^ xtime(column[0] ^ column[1])
    column[1] ^= all_xor ^ xtime(column[1] ^ column[2])
    column[2] ^= all_xor ^ xtime(column[2] ^ column[3])
    column[3] ^= all_xor ^ xtime(column_zero ^ column[3])


def mix_columns(state: VectorArray) -> None:
    """
    В рамках этой трансформации каждая колонка в State представляется в виде многочлена
    и перемножается в поле GF(2^8) по модулю x^4 + 1 с фиксированным многочленом 3x^3 + x^2 + x + 2.

    Добавляет диффузию в шифр.
    """
    for round_index in state:
        mix_column(round_index)


def state_from_bytes(data: bytes) -> VectorArray:
    """
    Массив байтов преобразуется в прямоугольный массив State.
    """
    state = [data[i * 4:(i + 1) * 4] for i in range(len(data) // 4)]
    return state  # type: ignore


def bytes_from_state(state: VectorArray) -> bytes:
    """
    State превращается в массив байтов.
    """
    return bytes(state[0] + state[1] + state[2] + state[3])


def sub_word(word: Vector) -> bytes:
    """
    Функция для перестановки байт.
    """
    substituted_word = bytes(STRAIGHT_BOX[i] for i in word)
    return substituted_word


def rcon(i: int) -> bytes:
    """
    Rcon — константная таблица, значения которой в шестнадцатеричной системе счисления, определена для преобразований.
    """
    # From Wikipedia
    rcon_lookup = bytearray.fromhex('01020408102040801b36')
    rcon_value = bytes([rcon_lookup[i - 1], 0, 0, 0])
    return rcon_value


def xor_bytes(array_a: bytes, array_b: bytes) -> bytes:
    """
    Вспомогательная функция для перемножения двух массивов байт.
    """
    return bytes([array_a_value ^ array_b_value for (array_a_value, array_b_value) in zip(array_a, array_b)])


def rot_word(word: Vector) -> Vector:
    """
    Вспомогательная функция.
    """
    return word[1:] + word[:1]


def key_expansion(key: bytes, nb: int = 4) -> VectorArrayArray:
    """

    State — промежуточный результат шифрования, который может быть представлен как прямоугольный массив байтов
        имеющий 4 строки и Nb колонок. Каждая ячейка State содержит значение размером в 1 байт
    Nb — число столбцов (32-х битных слов), составляющих State. Для стандарта регламентировано Nb = 4
    Nk — длина ключа в 32-х битных словах. Для AES, Nk = 4, 6, 8. Мы уже определились, что будем использовать Nk = 4
    Nr — количество раундов шифрования. В зависимости от длины ключа, Nr = 10, 12 или 14


    Эта вспомогательная трансформация формирует набор раундовых ключей — KeySchedule.

    KeySchedule представляет собой длинную таблицу, состоящую из Nb*(Nr + 1) столбцов или (Nr + 1) блоков,
    каждый из которых равен по размеру State.
    Первый раундовый ключ заполняется на основе секретного ключа, который вы придумаете, по формуле:

    KeySchedule[r][c] = SecretKey[r + 4c], r = 0,1...4; c = 0,1..Nk.
    """
    nk = len(key) // 4
    key_bit_length = len(key) * 8

    if key_bit_length == 128:
        nr = 10
    elif key_bit_length == 192:
        nr = 12
    else:  # 256-bit keys
        nr = 14

    w = state_from_bytes(key)

    for i in range(nk, nb * (nr + 1)):
        temp = w[i - 1]
        if i % nk == 0:
            temp = xor_bytes(sub_word(rot_word(temp)), rcon(i // nk))  # type: ignore
        elif nk > 6 and i % nk == 4:
            temp = sub_word(temp)  # type: ignore
        w.append(xor_bytes(w[i - nk], temp))  # type: ignore

    return [w[i * 4:(i + 1) * 4] for i in range(len(w) // 4)]


def aes_encryption(data: bytes, key: bytes) -> bytes:
    """
    Функция шифровки.
    """
    valid_aes_key(key)
    key_bit_length = len(key) * 8

    if key_bit_length == 128:
        nr = 10
    elif key_bit_length == 192:
        nr = 12
    else:  # 256-bit keys
        nr = 14

    state = state_from_bytes(data)

    key_schedule = key_expansion(key)

    add_round_key(state, key_schedule, round_number=0)

    for round_numer in range(1, nr):
        sub_bytes(state)
        shift_rows(state)
        mix_columns(state)
        add_round_key(state, key_schedule, round_numer)

    sub_bytes(state)
    shift_rows(state)
    add_round_key(state, key_schedule, round_number=nr)

    cipher = bytes_from_state(state)
    return cipher


def inv_shift_rows(state: VectorArray) -> None:
    """
    Обратная функция циклического сдвига.

    пример:
    [00, 10, 20, 30]     [00, 10, 20, 30]
    [01, 11, 21, 31] <-- [11, 21, 31, 01]
    [02, 12, 22, 32]     [22, 32, 02, 12]
    [03, 13, 23, 33]     [33, 03, 13, 23]
    """
    state[1][1], state[2][1], state[3][1], state[0][1] = state[0][1], state[1][1], state[2][1], state[3][1]
    state[2][2], state[3][2], state[0][2], state[1][2] = state[0][2], state[1][2], state[2][2], state[3][2]
    state[3][3], state[0][3], state[1][3], state[2][3] = state[0][3], state[1][3], state[2][3], state[3][3]


def inv_sub_bytes(state: VectorArray) -> None:
    """
    Обратная функция замены.
    """
    for round_index in range(len(state)):
        state[round_index] = [INVERSE_BOX[state[round_index][constant]] for constant in range(len(state[0]))]


def xtimes_0e(b: int) -> int:
    """
    Вспомогательная функция, которая выполняет умножение на соответствующую константу в GF(2^8).
    """
    # 0x0e = 14 = b1110 = ((x * 2 + x) * 2 + x) * 2
    return xtime(xtime(xtime(b) ^ b) ^ b)


def xtimes_0b(b: int) -> int:
    """
    Вспомогательная функция, которая выполняет умножение на соответствующую константу в GF(2^8).
    """
    # 0x0b = 11 = b1011 = ((x*2)*2+x)*2+x
    return xtime(xtime(xtime(b)) ^ b) ^ b


def xtimes_0d(b: int) -> int:
    """
    Вспомогательная функция, которая выполняет умножение на соответствующую константу в GF(2^8).
    """
    # 0x0d = 13 = b1101 = ((x*2+x)*2)*2+x
    return xtime(xtime(xtime(b) ^ b)) ^ b


def xtimes_09(b: int) -> int:
    """
    Вспомогательная функция, которая выполняет умножение на соответствующую константу в GF(2^8).
    """
    # 0x09 = 9  = b1001 = ((x*2)*2)*2+x
    return xtime(xtime(xtime(b))) ^ b


def inv_mix_column(column: Vector) -> None:
    """
    Обратная функция преобразования в многочлен.
    """
    column_zero, column_1, column_2, column_3 = column[0], column[1], column[2], column[3]
    column[0] = xtimes_0e(column_zero) ^ xtimes_0b(column_1) ^ xtimes_0d(column_2) ^ xtimes_09(column_3)
    column[1] = xtimes_09(column_zero) ^ xtimes_0e(column_1) ^ xtimes_0b(column_2) ^ xtimes_0d(column_3)
    column[2] = xtimes_0d(column_zero) ^ xtimes_09(column_1) ^ xtimes_0e(column_2) ^ xtimes_0b(column_3)
    column[3] = xtimes_0b(column_zero) ^ xtimes_0d(column_1) ^ xtimes_09(column_2) ^ xtimes_0e(column_3)


def inv_mix_columns(state: VectorArray) -> None:
    """
    Обратная функция преобразования в многочлен.
    """
    for r in state:
        inv_mix_column(r)


def inv_mix_column_optimized(column: Vector) -> None:
    """
    Обратная функция преобразования в многочлен. Оптимизированная.
    """
    u = xtime(xtime(column[0] ^ column[2]))
    v = xtime(xtime(column[1] ^ column[3]))
    column[0] ^= u
    column[1] ^= v
    column[2] ^= u
    column[3] ^= v


def inv_mix_columns_optimized(state: VectorArray) -> None:
    """
    Обратная функция преобразования в многочлен. Оптимизированная.
    """
    for r in state:
        inv_mix_column_optimized(r)
    mix_columns(state)


def aes_decryption(cipher: bytes, key: bytes) -> bytes:
    """
    Функция расшифровки.
    """
    valid_aes_key(key)
    key_byte_length = len(key)
    key_bit_length = key_byte_length * 8
    # nk = key_byte_length // 4

    if key_bit_length == 128:
        nr = 10
    elif key_bit_length == 192:
        nr = 12
    else:  # 256-bit keys
        nr = 14

    state = state_from_bytes(cipher)
    key_schedule = key_expansion(key)
    add_round_key(state, key_schedule, round_number=nr)

    for round_index in range(nr - 1, 0, -1):
        inv_shift_rows(state)
        inv_sub_bytes(state)
        add_round_key(state, key_schedule, round_index)
        inv_mix_columns(state)

    inv_shift_rows(state)
    inv_sub_bytes(state)
    add_round_key(state, key_schedule, round_number=0)

    plain = bytes_from_state(state)
    return plain


def aes_ecb_encryption(plain: bytes, key: bytes) -> bytes:
    """
    Функция шифровки набора блоков по 16 байт.
    """
    cipher: Iterable[Any] = []
    valid_aes_key(key)

    for j in range(len(plain) // AES_BLOCK_SIZE):
        p_j = plain[j * AES_BLOCK_SIZE:(j + 1) * AES_BLOCK_SIZE]
        c_j = aes_encryption(p_j, key)
        cipher += c_j  # type: ignore
    return bytes(cipher)


def aes_ecb_decryption(cipher: bytes, key: bytes) -> bytes:
    """
    Функция расшифровки набора блоков по 16 байт.
    """
    plain: Iterable[Any] = list()
    valid_aes_key(key)

    for j in range(len(cipher) // AES_BLOCK_SIZE):
        c_j = cipher[j * AES_BLOCK_SIZE:(j + 1) * AES_BLOCK_SIZE]
        p_j = aes_decryption(c_j, key)
        plain += p_j  # type: ignore
    return bytes(plain)



from typing import Dict, List, Any
from datetime import datetime, tzinfo

from data_aggregator_sdk.constants.enums import IntegrationV0MessageEvent
from data_aggregator_sdk.integration_message import IntegrationV0MessageData


from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet


# PACKET (64 bits)   water5_device_warm_reset_v_fluo_a
#
# RESULT int:        188978561091
# RESULT bin:  MSB   0000000000000000000000000010110000000000000000000000000001000011   LSB
# RESULT hex:  LE    43 00 00 00 2c 00 00 00
#
#
# name            type  size  value(int)                                                        data(bits)
# --------------------------------------------------------------------------------------------------------
# pack_id         u8       8          67                                                          01000011
# UNUSED          -       24           0                                  000000000000000000000000
# UNUSED          -        2           0                                00
# pin_reset       bool     1           1                               1
# low_voltage     bool     1           1                              1
# software_reset  bool     1           0                             0
# watchdog_reset  bool     1           1                            1
# RESERVED        u26     26           0  00000000000000000000000000


class Water5DeviceWarmResetVFluoAData(Packet):
    pin_reset: bool
    low_voltage: bool
    software_reset: bool
    watchdog_reset: bool

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((67) & (2 ** (8) - 1)) << size
        size += 8
        result |= ((0) & (2 ** (24) - 1)) << size
        size += 24
        result |= ((0) & (2 ** (2) - 1)) << size
        size += 2
        assert isinstance(data.pin_reset, bool)
        result |= ((int(data.pin_reset)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.low_voltage, bool)
        result |= ((int(data.low_voltage)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.software_reset, bool)
        result |= ((int(data.software_reset)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.watchdog_reset, bool)
        result |= ((int(data.watchdog_reset)) & (2 ** (1) - 1)) << size
        size += 1
        return result.to_bytes(8, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'Water5DeviceWarmResetVFluoAData':
        result__el_tmp1: Dict[str, Any] = dict()
        if 67 != buf.shift(8):
            raise ValueError("pack_id: buffer doesn't match value")
        buf.shift(24)
        buf.shift(2)
        result__el_tmp1["pin_reset"] = bool(buf.shift(1))
        result__el_tmp1["low_voltage"] = bool(buf.shift(1))
        result__el_tmp1["software_reset"] = bool(buf.shift(1))
        result__el_tmp1["watchdog_reset"] = bool(buf.shift(1))
        buf.shift(26)
        result = Water5DeviceWarmResetVFluoAData(**result__el_tmp1)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        return [
            IntegrationV0MessageData(
                dt=received_at,
                events=[
                    *([] if any((self.pin_reset, self.low_voltage, self.software_reset, self.watchdog_reset)) else [IntegrationV0MessageEvent.RESET]),
                    *([] if not self.pin_reset else [IntegrationV0MessageEvent.RESET_PIN]),
                    *([] if not self.low_voltage else [IntegrationV0MessageEvent.RESET_LOW_VOLTAGE]),
                    *([] if not self.software_reset else [IntegrationV0MessageEvent.RESET_SOFTWARE]),
                    *([] if not self.watchdog_reset else [IntegrationV0MessageEvent.RESET_WATCHDOG]),
                ],
            ),
        ]



from decimal import Decimal
from typing import List, Any, Dict
from datetime import datetime, tzinfo
from data_aggregator_sdk.constants.enums import IntegrationV0MessageEvent, SensorType
from data_aggregator_sdk.integration_message import IntegrationV0MessageData, \
    IntegrationV0MessageConsumption, CounterType, ResourceType, IntegrationV0MessageSensor

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet


# PACKET (64 bits)   smpm_ul_device_jupiter_08b_counter_ch1
#
# RESULT int:        31139818566057830
# RESULT bin:  MSB   00000000 01101110 10100001 01111111 11111111 11111111 11111111 01100110   LSB
# RESULT hex:  LE    66 ff ff ff 7f a1 6e 00
#
# name                     type   size  value(int)                                                        data(bits)
# -----------------------  -----  ----  ----------  ----------------------------------------------------------------
# packet_type_id.0.VAL     u7        7         102                                                           1100110
# packet_type_id.0.DFF     bool      1           0                                                          0
# value                    u32      32  2147483647                          01111111111111111111111111111111
# battery_volts            uf6p1     6          33                    100001
# temperature              u7        7          58             0111010
# event_reset              bool      1           1            1
# event_low_battery_level  bool      1           1           1
# RESERVED                 u9        9           0  000000000


class SmpmUlDeviceJupiter08BCounterCh1Data(Packet):
    value: int
    battery_volts: float
    temperature: int
    event_reset: bool
    event_low_battery_level: bool

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((102) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.value, int)
        result |= (((data.value) & 4294967295) & (2 ** (32) - 1)) << size
        size += 32
        assert isinstance(data.battery_volts, (int, float))
        assert 0.0 <= data.battery_volts <= 6.3
        result |= ((int(round(float(data.battery_volts) * 10.0, 0))) & (2 ** (6) - 1)) << size
        size += 6
        assert isinstance(data.temperature, int)
        assert -35 <= data.temperature <= 92
        result |= (((data.temperature + 35)) & (2 ** (7) - 1)) << size
        size += 7
        assert isinstance(data.event_reset, bool)
        result |= ((int(data.event_reset)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_low_battery_level, bool)
        result |= ((int(data.event_low_battery_level)) & (2 ** (1) - 1)) << size
        size += 1
        return result.to_bytes(8, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceJupiter08BCounterCh1Data':
        result__el_tmp1: Dict[str, Any] = dict()
        if 102 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp1["value"] = buf.shift(32) + 0
        result__el_tmp1["battery_volts"] = round(buf.shift(6) / 10.0, 1)
        result__el_tmp1["temperature"] = buf.shift(7) + -35
        result__el_tmp1["event_reset"] = bool(buf.shift(1))
        result__el_tmp1["event_low_battery_level"] = bool(buf.shift(1))
        result = SmpmUlDeviceJupiter08BCounterCh1Data(**result__el_tmp1)
        buf.shift(9)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        return [
            IntegrationV0MessageData(
                dt=received_at,
                sensors=[
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.battery_volts)),
                        sensor_type=SensorType.BATTERY,
                    ),
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.temperature)),
                        sensor_type=SensorType.TEMPERATURE,
                    ),
                ],
                consumption=[
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        channel=1,
                        value=Decimal(str(self.value)),
                        overloading_value=Decimal(str(1099511627775.0)),
                    ),
                ],
                events=[
                    *([IntegrationV0MessageEvent.BATTERY_IS_LOW] if self.event_low_battery_level else []),
                    *([IntegrationV0MessageEvent.RESET] if self.event_reset else []),
                ],
            ),
        ]
